﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ATH
{
    public class EntityHandler
    {
        private Entity[] entities;
        private List<SimClass> simClasses;

        public EntityHandler()
        {
            this.simClasses = new List<SimClass>();
            this.CreateNumericSpace();
        }

        private void CreateNumericSpace()
        {
            string[] row;
            Console.WriteLine("Creating Numeric space...");
            DataCSVHandler.ReadFile();
            List<Entity> e = new List<Entity>();
            while((row = DataCSVHandler.GetRow()) != null)
            {
                e.Add(Entity.Create(row));
            }
            this.entities = e.ToArray();
        }
        
        public void DumpSimilaritySpace()
        {
            if (Config.DUMP_MATRIX) DataCSVHandler.OpenOutputFile();
            Stopper.GetLastElapsed("Creating" + ((Config.DUMP_MATRIX) ? " and dumping" : "") + " similarity space started. ");
            for (int i = 0; i < this.entities.Length; i++)
            {
                Entity e1 = this.entities[i];
                for (int j = 0; j < this.entities.Length; j++)
                {
                    Entity e2 = this.entities[j];
                    float simValue = SimSpaceHelper.GetWeightedSimValue (e1, e2);
                    if (Config.DUMP_MATRIX) DataCSVHandler.AppendOutput(e1, e2, simValue);
                    e1.RowSim += simValue;
                    
                }
                if (i % 100 == 0)
                {
                    if (Config.DUMP_MATRIX) DataCSVHandler.WriteCache();
                    SimSpaceHelper.WritePercentageDone((this.entities.Length * i) / Math.Pow(this.entities.Length, 2));
                }
                //this.simMatrix.Add(SimilarityRow.CreateRow(e1, rowData));
            }
            if (Config.DUMP_MATRIX) DataCSVHandler.CloseOutputFile();
            SimSpaceHelper.WritePercentageDone(1);
            
        }
        public void GetClasses()
        {
            Stopper.GetLastElapsed("Calculating Similarity Classes started. ");
            int? index;
            do
            {
                index = GetRepresentative();
                if (index.HasValue)
                {
                    SimClass newClass = GetSimClass(index.Value);
                    if (newClass != null)
                    {
                        this.simClasses.Add(newClass);
                        Console.Write(this.simClasses.Count + " similarity classes found.               ");
                        Console.CursorLeft = 0;
                    }
                }
            } while (index.HasValue);
        }
        public SimClass GetSimClass(int repIndex)
        {
            SimClass newClass = new SimClass();
            newClass.SetRepresentative(this.entities[repIndex]);
            List<SimilarEntity> members = SimilarEntities(repIndex).ToList();
            foreach (var similarEntity in members)
            {
                newClass.AddEntity(this.entities[similarEntity.Index]);
                this.entities[similarEntity.Index].Classified = true;
            }
            return (newClass.Entities.Count > Config.MIN_SIMILARITY_CLASS_SIZE) ? newClass : null;
        }
        private int? GetRepresentative()
        {
            int? index = null;

            float maxValue = float.MinValue;
            List<Entity> candidates = this.entities.Where(ent => !this.entities[ent.Id].Classified).ToList();
            foreach (Entity entity in candidates)
            {
                if (entity.RowSim > maxValue)
                {
                    maxValue = entity.RowSim;
                    index = entity.Id;
                }
            }
            return index;
        }
        public IEnumerable<SimilarEntity> SimilarEntities(int entityIndex)
        {
            for (int i = 0; i < this.entities.Length; i++)
            {
                if (i == entityIndex)
                {
                    continue;
                }
                float similarity = SimSpaceHelper.GetWeightedSimValue(this.entities[entityIndex], this.entities[i]);
                if (similarity > Config.SIMILARITY_MARGIN)
                {
                    yield return new SimilarEntity(i, this.entities[i].Name, similarity);
                }
            }
        }
        public bool IsClassificationValid()
        {
            return this.GetRepresentedPercentage() >= Config.MIN_SIMILARITY_RATIO;
        }

        public float GetRepresentedPercentage()
        {
            List<Entity> allRepresented = new List<Entity>();
            foreach (SimClass sc in this.simClasses)
            {
                allRepresented.AddRange(sc.Entities);
            }

            return (float)allRepresented.Distinct().ToList().Count / (float)this.entities.Length;
        }

        public string GetRepresentationMessage()
        {
            string message = "";
            string p = Math.Round(this.GetRepresentedPercentage() * 100, Config.DECIMAL_PRECISION).ToString();
            if (this.IsClassificationValid())
            {
                message = "Classification is valid. Covered percentage of the similarity reduction space is : " + p + "%";
            }
            else
            {
                string min = (Config.MIN_SIMILARITY_RATIO * 100).ToString();
                message = "Classification is not valid. Not covers the minimum " + min + "% of the similarity reduction space.(" + p + "%)";
            }
            message += "\n";
            message += "There are total of " + this.simClasses.Count + " similarity classes." + "\n";
            message += "Space qualifier vector: <Omin, Umin, Cmin, DEVmin>=<" + Config.SIMILARITY_MARGIN + ", " + Config.MIN_SIMILARITY_CLASS_SIZE + ", " + Config.MIN_SIMILARITY_RATIO + ", " + Config.MAX_DEVIANT_MARGIN + ">\n\n";
            return message;
        }
        public void DoAttributeReduction()
        {
            string result = this.GetRepresentationMessage();
            foreach (SimClass sc in this.simClasses)
            {
                result += sc.ToString() + "\n";
            }
            DataCSVHandler.Write(result, SimClass.ClassesOutPath);
        }

        public List<SimClass> SimClasses
        {
            get { return simClasses; }
        }
    
    }
}
